{% extends "base.Dockerfile.jinja" %}

{% block install_requirements %}
COPY ./{{training_requirements_filename}} {{training_requirements_filename}}
RUN pip install -r {{training_requirements_filename}}  --no-cache-dir && rm -rf /root/.cache/pip
{{ super() }}
{% endblock %}

{% block app_copy %}
COPY ./training /app
# todo: put this behind a flag
# COPY ./{{ config.train.train_module_dir }} /app/train
COPY ./config.yaml /app/config.yaml
{% endblock %}


{% block run %}
CMD exec python3 /app/job.py
{% endblock %}
